from __future__ import annotations

from ase import Atoms

PF6 = Atoms(
    symbols="PF6",
    positions=[
        [6.747, 7.453, 7.469],
        [7.944, 6.319, 7.953],
        [6.127, 6.381, 6.461],
        [5.794, 8.645, 7.001],
        [5.815, 7.032, 8.699],
        [7.617, 8.534, 8.484],
        [7.91, 7.908, 6.284],
    ],
)  # the PF6 atoms type: ase.Atoms()

Li = Atoms(
    symbols="Li",
    positions=[[0, 0, 0]],
)  # the Li atoms type: ase.Atoms()


def dock_atoms(ship_atoms, dock_atoms=None, offset=1.5):
    # TODO: this dock function is not general enough, it should be able to dock any two atoms
    """
    Dock the ship🚢 atoms to the dock⚓ atoms (default is PF6).

    Parameters:
    -----------
    ship_atoms (ase.Atoms): The ship atoms.
    """
    if dock_atoms is None:
        dock_atoms = PF6.copy()
    ship_atoms = ship_atoms.copy()
    ship_atoms_center = ship_atoms.get_center_of_mass()
    ship_atoms_center[0] = max(ship_atoms.positions.T[0])
    dock_atoms_center = dock_atoms.get_center_of_mass()
    dock_atoms_center[0] = min(dock_atoms.positions.T[0])
    vector = ship_atoms_center - dock_atoms_center
    offset = [offset, 0, 0]
    dock_atoms.positions = dock_atoms.positions + vector + offset
    ship_atoms.extend(dock_atoms)
    return ship_atoms


def tmp_atoms(atoms, filename="tmp.xyz", create_tmp_folder=True):
    """
    Write the atoms to a temporary file in the tmp folder and return the path.

    Args:

        atoms (ase.Atoms): The atoms object.
        filename (str): The filename of the temporary file.

    Returns:

        filepath (str): The path of the temporary file
    """
    import os

    from ase.io import write

    _CWD = os.getcwd()
    if create_tmp_folder:
        from monty.os import makedirs_p

        makedirs_p(f"{_CWD}/tmp")
        filepath = os.path.join(f"{_CWD}/tmp", filename)
    else:
        filepath = os.path.join(_CWD, filename)
    write(filepath, atoms, format="xyz")
    return filepath



"""os utilities for himatcal"""

import os
import re

from monty.os import makedirs_p


def labeled_dir(main_workdir, label):
    """
    Create a new folder in the main_workdir with the label.

    Args:

        main_workdir (str): The main working directory.
        label (str): The label of the folder.

    Returns:

        folder_path (str): The path of the new folder.
    """
    # Get the folder names in main_workdir
    folder_names = [
        name
        for name in os.listdir(main_workdir)
        if os.path.isdir(os.path.join(main_workdir, name))
    ]
    numbers = [
        int(re.search(r"\d+", name).group())
        for name in folder_names
        if re.search(r"\d+", name)
    ]
    new_number = max(numbers) + 1 if numbers else 1
    # Create new folder
    folder_name = f"{new_number:02d}.{label}"
    folder_path = os.path.join(main_workdir, folder_name)
    makedirs_p(folder_path)
    print(f"Created new folder: {folder_path}")
    return folder_path


def get_chg_mult(molname):
    """Get the label, chg and mult from the name of a molecule, format: {label}-c{charge}s{mult}"""
    pattern = r"(.*?)-c(n?\d)s(\d+)"
    match = re.match(pattern, molname)
    if match:
        label, chg, mult = match.groups()
        chg = "-" + chg[1:] if chg.startswith("n") else chg
        return label, int(chg), int(mult)
    else:
        return None, None, None
    
def write_chg_mult_label(label, chg, mult):
    """Write the label, chg and mult to a string, format: {label}-c{charge}s{mult}"""
    if chg < 0:
        chg = f"n{abs(chg)}"
    return f"{label}-c{chg}s{mult}"



# view the slab using py3Dmol
import py3Dmol
from ase.io import read, write

def view_atoms(atoms, format='xyz'):
    """
    View the atoms in jupyter notebook using py3Dmol.
    """
    write('tmp_atoms', atoms, format=format)
    atoms_data = open('tmp_atoms', 'r').read()
    view = py3Dmol.view(width=800, height=400)
    view.addModel(atoms_data, format)
    view.setStyle({'stick':{}})
    view.zoomTo()
    return view

def show_xyz_mol(xyz_file):
    """
    Visualize a stk molecule using py3Dmol.
    """
    mol = open(xyz_file).read()
    p = py3Dmol.view(
        data = mol,
        style = {'stick': {'colorscheme': 'Jmol'}},
        width=400,
        height=400,
    )
    p.setBackgroundColor('white')
    p.zoomTo()
    p.show()

def xyz_to_mol(xyz_file, write_mol=True):
    """
    Convert a xyz file to a mol file and block.
    """
    from openbabel import pybel as pb
    mol = next(pb.readfile('xyz', xyz_file))
    if write_mol:
        mol.write('mol', f'{xyz_file}.mol', overwrite=True)
        return open(f'{xyz_file}.mol').read()

# * Gasiteger charge visualization
def plot_gasteiger_charges(mol):
    """
    Plot Gasteiger charges on a molecule.
    """
    from rdkit.Chem.Draw import SimilarityMaps
    from rdkit.Chem import AllChem
    AllChem.ComputeGasteigerCharges(mol)
    contribs = [float(mol.GetAtomWithIdx(i).GetProp('_GasteigerCharge')) for i in range(mol.GetNumAtoms())]
    SimilarityMaps.GetSimilarityMapFromWeights(
        mol, contribs, colorMap='jet', contourLines=10
    )


"""Init data for himatcal package."""

from __future__ import annotations

from importlib.metadata import version

from himatcal.settings import HimatcalSettings

# load the version from the pyproject.toml file
__version__ = version("himatcal")

# load the settings
SETTINGS = HimatcalSettings()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray


def add(a: float, b: float) -> float:
    """
    A function that adds two numbers.

    Parameters
    ----------
    a
        First number to add.
    b
        Second number to add.

    Returns
    -------
    float
        The sum of a and b.
    """
    return a + b


def divide(a: float, b: float) -> float:
    """
    A function that divides two numbers, i.e. a/b.

    Parameters
    ----------
    a
        The numerator
    b
        The denominator

    Returns
    -------
    float
        The value for a/b
    """
    if b == 0:
        raise ValueError("Uh oh! The value for b should not be 0.")

    return a / b


def make_array(val: float, length: int = 3) -> NDArray:
    """
    A function to transform a number into a numpy array.

    Parameters
    ----------
    val
        Number to turn into an array.
    length
        The length of the array.

    Returns
    -------
    NDArray
        An array composed of `val`.
    """
    return np.array([val] * length)



from typing import Optional

import typer
from rich import print
from typing_extensions import Annotated

app = typer.Typer(
    name="himatcal",
    help="A CLI for Himatcal project",
    no_args_is_help=True,
    rich_markup_mode="rich",
)


def machine_name():
    """
    Returns a list of machine names.

    Returns:
        list: A list of machine names.
    """
    return ["local", "sghpc1", "yeesuan", "XMU"]


@app.command("hello", help="Command to say hello :red_heart-emoji:")
def hello(name: Annotated[Optional[str], typer.Argument(help="test command")] = None):
    """
    test command
    """
    if name is not None:
        print(f"Hello, {name}!")
        return f"Hello, {name}!"
    else:
        print("Hello, world!")


@app.command(
    "sub_gs",
    help="Submit a [bold green]Gaussian[/bold green] job",
    no_args_is_help=True,
)
def sub_gs(
    input_file: Annotated[str, typer.Argument(help="Gaussian input file.")],
    label: Annotated[str, typer.Option(help="label for the calculation")] = "Gaussian",
    machine: Annotated[
        str,
        typer.Option(
            help="Machine to run the job, select 'local' and 'sghpc1'.",
            autocompletion=machine_name,
        ),
    ] = "sghpc1",
):
    """
    Submit a Gaussian job
    """
    from himatcal.recipes.gaussian.flow import sub_gs

    dispatch_id = sub_gs(input_file, label, machine)
    print(f"Dispatched job with id: {dispatch_id}")


@app.command("gs", help="Submit a Gaussian job", no_args_is_help=True)
def gs_jobs(
    inputfile: Annotated[str, typer.Argument(help="file as input")],
    command: Annotated[
        str, typer.Argument(help="Command to submit the job")
    ] = "sub_gs",
):
    """
    Submit a Gaussian job
    """
    if command == "sub_gs":
        sub_gs(input_file=inputfile, label="Gaussian", machine="sghpc1")
    elif command == "relax":
        from ase.io import read
        from himatcal.recipes.gaussian.core import relax_job

        atoms = read(inputfile)
        relax_job(
            atoms=atoms,
            charge=0,
            spin_multiplicity=1,
            label="relax",
            xc="B3LYP",
            basis="6-31+g(d) em=GD3BJ",
            freq=False,
        )


@app.command(
    "ts_opt", help="Submit a transition state optimization job", no_args_is_help=True
)
def ts_opt(
    atoms_path: Annotated[str, typer.Argument(help="Atoms to optimize")],
    charge: Annotated[int, typer.Argument(help="Charge of the molecule")],
    mult: Annotated[int, typer.Argument(help="Spin multiplicity of the molecule")],
    label: Annotated[str, typer.Option(help="label for the calculation")] = "TS",
    machine: Annotated[
        str,
        typer.Option(
            help="Machine to run the job, select 'local' and 'sghpc1'.",
            autocompletion=machine_name,
        ),
    ] = "sghpc1",
):
    """
    Submit a transition state optimization job
    """
    from himatcal.recipes.gaussian.flow import ts_opt

    dispatch_id = ts_opt(atoms_path, charge, mult, label, machine)
    print(f"Dispatched job with id: {dispatch_id}")


@app.command("irc", help="Submit a Gaussian IRC calculation", no_args_is_help=True)
def irc(
    atoms_path: Annotated[str, typer.Argument(help="Atoms to optimize")],
    chg: Annotated[int, typer.Argument(help="Charge of the molecule")],
    mult: Annotated[int, typer.Argument(help="Spin multiplicity of the molecule")],
    label: Annotated[str, typer.Option(help="label for the calculation")] = "IRC",
    machine: Annotated[
        str,
        typer.Option(
            help="Machine to run the job, select 'local' and 'sghpc1'.",
            autocompletion=machine_name,
        ),
    ] = "sghpc1",
):
    """
    Submit a Gaussian IRC calculation
    """
    from himatcal.recipes.gaussian.flow import irc_flow

    dispatch_id = irc_flow(atoms_path, chg, mult, label, machine)
    print(f"Dispatched job with id: {dispatch_id}")


@app.command(
    "ts_flow", help="Submit a transition state optimization job", no_args_is_help=True
)
def ts_flow(
    atoms_path: Annotated[str, typer.Argument(help="Atoms to optimize")],
    chg: Annotated[int, typer.Argument(help="Charge of the molecule")],
    mult: Annotated[int, typer.Argument(help="Spin multiplicity of the molecule")],
    label: Annotated[str, typer.Option(help="label for the calculation")] = "TS-flow",
    machine: Annotated[
        str,
        typer.Option(
            help="Machine to run the job, select 'local' and 'sghpc1'.",
            autocompletion=machine_name,
        ),
    ] = "sghpc1",
):
    """
    Submit a transition state optimization job
    """
    from himatcal.recipes.gaussian.flow import ts_flow

    dispatch_id = ts_flow(atoms_path, chg, mult, label)
    print(f"Dispatched job with id: {dispatch_id}")


@app.command(
    "relax", help="Submit a Gaussian relaxation calculation", no_args_is_help=True
)
def relax(
    atoms_path: Annotated[str, typer.Argument(help="The path of atoms file to relax")],
    chg: Annotated[int, typer.Argument(help="Charge of the molecule")],
    mult: Annotated[int, typer.Argument(help="Spin multiplicity of the molecule")],
    label: Annotated[str, typer.Option(help="label for the calculation")] = "relax",
    machine: Annotated[
        str,
        typer.Option(
            help="Machine to run the job, select 'local' and 'sghpc1'.",
            autocompletion=machine_name,
        ),
    ] = "sghpc1",
):
    """
    Submit a Gaussian relaxation calculation
    """
    from himatcal.recipes.gaussian.flow import relax

    dispatch_id = relax(atoms_path, chg, mult, label, machine)
    print(f"Dispatched job with id: {dispatch_id}")


@app.command("cancel", help="Cancel a covalent job", no_args_is_help=True)
def cancel_job(job_id: Annotated[str, typer.Argument(help="Job id to cancel")]):
    """
    Cancel a covalent job
    """
    import covalent as ct

    ct.cancel(job_id)  # TODO: seems like this is not working
    print(f"Job with id {job_id} cancelled successfully")


@app.command("test_exec")

@app.command("status", help="Get the status of a covalent job", no_args_is_help=True)
def job_status(job_id: Annotated[str, typer.Argument(help="Job id to get status")]):
    """
    Get the status of a covalent job
    """
    import covalent as ct

    result = ct.get_result(job_id)
    status = result.status
    print(f"Job with id {job_id} is {status}")
    return status


@app.command("redispatch", help="Redispatch a covalent job", no_args_is_help=True)
def redispatch_job(
    job_id: Annotated[str, typer.Argument(help="Job id to redispatch")],
    reuse: Annotated[bool, typer.Option(help="Reuse the previous job")] = False,
):
    """
    Redispatch a covalent job
    """
    import covalent as ct

    redispatch_func = ct.redispatch(
        job_id,
        reuse_previous_results=reuse,
    )
    redispatch_id = redispatch_func()
    print(f"Job redispathed to {redispatch_id} (reuse_previous_results={reuse})")


@app.command(
    "extract_result", help="Extract the result of a covalent job", no_args_is_help=True
)
def extrcat_result(
    job_id: Annotated[str, typer.Argument(help="Job id to extract result")],
):
    """
    Extract the result of a covalent job
    """
    from himatcal.utils.ct import extract_result

    results = extract_result(job_id)
    return results


@app.command(
    "cas", help="Get molecular structure from CAS number", no_args_is_help=True
)
def cas(
    cas: Annotated[str, typer.Argument(help="CAS number")],
    write: Annotated[str, typer.Option(help="Write the structure to file")] = "xyz",
):
    """
    Get molecular structure from CAS number
    """
    from himatcal.recipes.mol.core import get_molecular_structure

    atoms = get_molecular_structure(molecular_cas=cas, write_to_file=write)
    print(atoms)


if __name__ == "__main__":
    app()



"""Settings for himatcal"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

if TYPE_CHECKING:
    from typing import Any

_DEFAULT_CONFIG_FILE_PATH = Path("~", ".himatcal.yaml").expanduser().resolve()


class HimatcalSettings(BaseSettings):
    """
    Settings for himatcal
    """

    model_config = SettingsConfigDict(
        env_prefix="himatcal_",
        env_nested_delimiter="__",
        env_parse_none_str="None",
        extra="forbid",
        validate_assignment=True,
    )

    CONFIG_FILE: Path = Field(
        _DEFAULT_CONFIG_FILE_PATH, description=("Path to the YAML config file")
    )

    # ---------------------------
    # external software Settings
    # ---------------------------
    XTB_EXE_PATH: Path = Field(Path(), description=("Path to the xtb executable"))

    CREST_EXE_PATH: Path = Field(Path(), description=("Path to the crest executable"))

    CREST_EXE_PATH_V3: Path = Field(
        Path(), description=("Path to the crest executable")
    )  # TODO: test the difference between CREST_EXE_PATH and CREST_EXE_PATH_V3

    ISOSTAT_EXE_PATH: Path = Field(
        Path(), description=("Path to the isostat executable")
    )

    FORMCHK_PATH: Path = Field(Path(), description=("Path to the formchk executable"))

    MULTIWFN_PATH: Path = Field(Path(), description=("Path to the Multiwfn executable"))

    SOBTOP_PATH: Path = Field(Path(), description=("Path to the sobtop executable"))

    OBABEL_PATH: Path = Field(Path(), description=("Path to the obabel executable"))

    GROMACS_PATH: Path = Field(Path(), description=("Path to the gromacs executable"))

    # ---------------------------
    # external modle Settings
    # ---------------------------
    MABASIS_PATH: Path = Field(Path(), description=("Path to the mabasis"))

    AIMNET2_MODEL_PATH: Path = Field(Path(), description=("Path to the aimnet2 model"))

    # ---------------------------
    # external api Settings
    # ---------------------------

    MAPI_KEY: str = Field("", description=("API key for the Materials Project"))

    CHEMSPIDER_API_KEY: str = Field("", description=("API key for ChemSpider"))

    # ---------------------------
    # external key file
    # ---------------------------
    SGHPC1_KEY_PATH: Path = Field("", description=("Path to the ssh key file of sghpc1"))
    YEESUAN_KEY_PATH: Path = Field("", description=("Path to the ssh key file of yeesuan"))
    XMU_KEY_PATH: Path = Field("", description=("Path to the ssh key file of XMU"))
    
    # ---------------------------
    # MongoDB URI
    # ---------------------------
    MONGODB_URI: str = Field(
        "", description=("MongoDB URI for the database connection")
    )

    @model_validator(mode="before")
    @classmethod
    def load_user_setting(cls, settings: dict[str, Any]) -> dict[str, Any]:
        """Load user settings"""
        return _type_handler(_use_custom_config_settings(settings))


def _use_custom_config_settings(settings: dict[str, Any]) -> dict[str, Any]:
    """Use custom settings from the config file"""
    from monty.serialization import loadfn

    config_file_path = (
        Path(settings.get("CONFIG_FILE", _DEFAULT_CONFIG_FILE_PATH))
        .expanduser()
        .resolve()
    )

    new_settings = {}  # type: dict
    if config_file_path.exists() and config_file_path.stat().st_size > 0:
        new_settings |= loadfn(config_file_path)

    new_settings.update(settings)
    return new_settings


def _type_handler(settings: dict[str, Any]) -> dict[str, Any]:
    """
    Convert common strings to their proper types.

    Parameters
    ----------
    settings : dict
        Initial settings.

    Returns
    -------
    dict
        Updated settings.
    """
    for key, value in settings.items():
        if isinstance(value, str):
            if value.lower() in {"null", "none"}:
                settings[key] = None
            elif value.lower() in {"true", "false"}:
                settings[key] = value.lower() == "true"

    return settings
